/*
	control.cpp
	   ,      
	   .
	     
	SysTick_Handler,    
	 ( ,    
	,       
	    ).

	      
	display.h, display.cpp
*/

#include "stm32f10x.h"
#include "display.h"
#include <limits>

//  ,     .
#define ND 8

//    , ,  
//    .
enum UPDATE_MASK{	//  :
	F_UPDATE=1,		// ;
	H_UPDATE=2,		//    (   Hold);
	D_UPDATE=4,		//     H;
	T_UPDATE=8,		// ;
	I_UPDATE=16,	//   .
	ALL_UPDATE=31,	//    .
	F_OFF=32		//  F (  ).
};

//    fm_start, get_result,   
//   (    main.cpp):

//   (    ).
// timer_div1, timer_div2   ,
//   timer_div1*timer_div2/SystemCoreClock, 
// timer_div1=1..65536; timer_div2=2..65536.
// k -      , ,
//    
// k*timer_div1*timer_div2/SystemCoreClock
//      ,  ,
//  false.
bool fm_start(uint32_t timer_div1, uint32_t timer_div2, uint32_t k);

//       .
//    ,   :
// m -       ;
// n -     ( ) 
//  ;
// cnt -   ;
// c0 -   32-    -
//  .
//   false    (  
//  -        ).
bool get_result(uint32_t &m, uint32_t &n, uint32_t &cnt, uint32_t &c0);

//  ( ),    .

//   :
// 0 -      (,   );
// 1 -     .
int view_mode=0;

//     interval_tbl,  
//  .
//    delta*interval_tbl[interval_mode]
int interval_mode=0;

//      .
const int interval_tbl[]={2, 5, 10};
const float delta=0.1;

//   , .
#define UPDATE_INTERVAL 400

//    (   ), .
#define BLINK_HALF_PERIOD 200

//        .
class Info
{
public:
	float F;		// .
	float H;		//       Hold.
	float D;		//  F  H.
	float T;		// ,   1/F.
	float tau;		//   .
	bool blink;		// true,   F  .
};

//      .
Info dinfo={0, 0, 0, 0, 0, true};

//     "".
void keyboard_init()
{
	//     /.
	RCC->APB2ENR|=
			RCC_APB2ENR_IOPCEN|
			RCC_APB2ENR_IOPAEN;
	//  ,      
	//   (PA0    ; PC0  PC1 -
	//  ).
	GPIOC->CRL=
			GPIOC->CRL&~0x000000FF|
						0x00000088;
	GPIOA->CRL=
			GPIOA->CRL&~0x0000000F|
						0x00000008;
	GPIOC->BSRR=GPIO_BSRR_BS0|GPIO_BSRR_BS1;
	GPIOA->BRR=GPIO_BRR_BR0;
}

//      "Interval";
//  true,   .
bool get_btn_interval_state()
{
	return !(GPIOC->IDR&GPIO_IDR_IDR0);
}

//      "Hold";
//  true,   .
bool get_btn_hold_state()
{
	return !(GPIOC->IDR&GPIO_IDR_IDR1);
}

//      "Mode";
//  true,   .
bool get_btn_mode_state()
{
	return GPIOA->IDR&GPIO_IDR_IDR0;
}

//       
// ;  0   - .
float get_frequency()
{
	uint32_t m=0;
	uint32_t n=0;
	uint32_t cnt=0;
	uint32_t c0;

	float f=0;
	get_result(m, n, cnt, c0);
	if(n!=0)
		f = 8 * (float)SystemCoreClock * m / n;
	return f;
}

//       dinfo, 
//    (   , 
//       mask).
void display_update(int update_mask)
{
	//       ;
	//      .
	static float F_prev=0;
	static float T_prev=0;
	static float D_prev=0;

	if(update_mask&F_OFF)
		display_print_blank(F_ROW, 2);
	if(update_mask&F_UPDATE)
	{
		int xp=(view_mode==0)?get_xp(dinfo.F, F_prev):0;
		display_print(F_ROW, 2, dinfo.F, ND, xp, "Hz", false);
		F_prev=dinfo.F;
	}

	if(update_mask&H_UPDATE)
	{
		display_print(H_ROW, 2, dinfo.H, ND,
				(view_mode==0)?get_xp(dinfo.H):0,
						"Hz", false);
	}

	//    D,  , 
	// ,    D,    D.
	//        
	// .
	if(update_mask&D_UPDATE)
	{
		uint32_t nd=ND;
		int xp=0;

		if(view_mode==0)
		{
			xp=get_xp(dinfo.D, D_prev);

			float ad=fabsf(dinfo.D);
			float ff=(dinfo.F>=dinfo.H)?dinfo.F:dinfo.H;
			if(ff!=0&&ff<1e10*ad)
				nd-=floorf(log10(ff/ad));
			if(nd<2)
				nd=2;
			if(nd>ND)
				nd=ND;
		}

		display_print(D_ROW, 2, dinfo.D, nd, xp, "Hz", true);
		D_prev=dinfo.D;
	}

	if(update_mask&T_UPDATE)
	{
		display_print(T_ROW, 2, dinfo.T, ND, get_xp(dinfo.T, T_prev), "s", false);
		T_prev=dinfo.T;
	}

	if(update_mask&I_UPDATE)
		display_print(I_ROW, 2, interval_tbl[interval_mode]*delta, 2, 0, "s", false);
}

//     .
//     ( 
// );      
//  .
//      main   main.cpp;
//     1.
extern "C" void SysTick_Handler()
{
	//  .
	static uint32_t ticks=0;

	//    (   
	//  ).
	static bool btn_interval_state=false;
	static bool btn_hold_state=false;
	static bool btn_mode_state=false;

	// true,       (  
	//       ).
	static bool first_call=true;

	int update_mask=0;

	if(first_call)
		update_mask=ALL_UPDATE;
	first_call=false;

	//     32 -    
	//   ""    
	// ,   ""  .
	//  Reset   ,   
	//  .
	if(!(ticks&0x1F))
	{
		//   Interval.
		bool btn_interval_state2=get_btn_interval_state();
		if(!btn_interval_state&&btn_interval_state2)
		{
			interval_mode++;
			if(interval_mode>=sizeof interval_tbl/sizeof interval_tbl[0])
				interval_mode=0;
			dinfo.tau=delta*interval_tbl[interval_mode];
			fm_start(240, 10000, interval_tbl[interval_mode]);
			update_mask|=I_UPDATE;
		}
		btn_interval_state=btn_interval_state2;

		//   Hold.
		bool btn_hold_state2=get_btn_hold_state();
		if(!btn_hold_state&&btn_hold_state2)
		{
			float H=get_frequency();
			if(dinfo.H!=H)
			{
				dinfo.H=H;
				update_mask|=H_UPDATE;
			}
			float D=dinfo.F-H;
			if(dinfo.D!=D)
			{
				dinfo.D=D;
				update_mask|=D_UPDATE;
			}
		}
		btn_hold_state=btn_hold_state2;

		//   State.
		bool btn_mode_state2=get_btn_mode_state();
		if(!btn_mode_state&&btn_mode_state2)
		{
			(++view_mode)&=1;
			update_mask=ALL_UPDATE;
		}
		btn_mode_state=btn_mode_state2;
	}

	//     .
	if(ticks%UPDATE_INTERVAL==0)
	{
		float F=get_frequency();
		if(dinfo.F!=F)
		{
			dinfo.F=F;
			update_mask|=F_UPDATE;
			dinfo.D=F-dinfo.H;
			update_mask|=D_UPDATE;
			dinfo.T=(dinfo.F>0)?1/dinfo.F:std::numeric_limits<float>::infinity();
			update_mask|=T_UPDATE;
		}
	}

	//         
	//   ( ).
	if(ticks%BLINK_HALF_PERIOD==0)
	{
		bool blink=dinfo.F==0;
		if(dinfo.blink&&!blink)
			update_mask|=F_UPDATE;
		dinfo.blink=blink;
		if(blink)
		{
			if(ticks%(2*BLINK_HALF_PERIOD)==0)
				update_mask|=F_UPDATE;
			else
			{
				update_mask&=~F_UPDATE;
				update_mask|=F_OFF;
			}
		}
	}

	//  ,        .
	if(update_mask)
		display_update(update_mask);

	ticks++;
}
