/*
      
 LCD 1.77" 160*128 (RGB)   ST7735s;
      ,  
     4-  
  ( SPI).
*/

/* 
*****     MCU STM32F100RB  *****
--------------------------------------------
  LCD	
--------------------------------------------
1. GND					GND
2. VCC					+3V3
3. SCK					PA5 (SPI1_SCK)
4. SDA					PA6 (SPI1_MISO), PA7 (SPI_MOSI)
5. RES					PB7 (*)
6. RS					PB6 (*)
7. CS					PA4 (*)
8. LEDA					(<=3V3)
--------------------------------------------
   LCD  SPI1 ,  
  PA5, PA6, PA7   
 LCD.
 (*)  (PA4, PB6, PB7)     
 .
*/

/*
      LCD.   
      RAM,     
 MCU    .     LCD  
 .
 
    ,    
 ,       (  DMA).    
     RAM,      
 .
   -     
        
      .
*/

#pragma once

#include "stm32f10x.h"
#include <math.h>

//      8*8.
#include <font8x8.h>

//    ,   R, G, B.
//  6     
// ,    .

//    .
#define COLOR_RED		0x000000FF
#define COLOR_GREEN		0x0000FF00
#define COLOR_BLUE		0x00FF0000
#define COLOR_YELLOW	0x0000FFFF
#define COLOR_WHITE		0x00FFFFFF
#define COLOR_BLACK		0x00000000

//       .
#define X_START		2
#define X_STOP		129
#define Y_START		1
#define Y_STOP		160

//        .
#define DISPLAY_WIDTH	(X_STOP-X_START+1)
#define DISPLAY_HIGHT	(Y_STOP-Y_START+1)

//       MCU STM32F100,
//    LCD.
#define IN_ANALOG		0x0
#define IN_FLOATING		0x4
#define ALT_PP_OUT_HS	0xB
#define MAIN_PP_OUT_HS	0x3

//     us ,  
// TIM16 (    st7735s.cpp).
void udelay(uint32_t us);

//       GPIO .
class Pin
{
public:
	GPIO_TypeDef *gpio;		//  (GPIOA, GPIOB, ...).
	uint16_t number;		//   (0, 1, 2, ..., 15).

	//       .
	void set(bool state)
	{
		uint16_t d=(state)?0:16;
		d+=number;
		gpio->BSRR=1<<d;
	}

	//     .
	bool read_in()
	{
		return gpio->IDR&(1<<number);
	}

	//     (  ODR).
	bool read_out()
	{
		return gpio->ODR&(1<<number);
	}

	//    (  4-
	// ).
	void config(int cfg)
	{
		cfg&=0xF;
		volatile uint32_t *cr=(number<8)?
				&gpio->CRL:
				&gpio->CRH;
		uint16_t d=(number&7)*4;
		*cr=*cr&~(0xF<<d)|(cfg<<d);
	}

	//     (  
	// 4- ).
	int get_config()
	{
		volatile uint32_t *cr=(number<8)?
				&gpio->CRL:
				&gpio->CRH;
		uint16_t d=(number&7)*4;
		return (*cr>>d)&0xF;
	}
};

//       .
//   SPI1;    
//  MCU   LCD.
class DisplayInterface
{
public:
	Pin spi_sck;
	Pin spi_miso;
	Pin spi_mosi;
	Pin nss;
	Pin res;
	Pin dcx;

	//   , SPI1  DMA.
	void init()
	{
		RCC->APB2ENR|=
					RCC_APB2ENR_IOPAEN|
					RCC_APB2ENR_IOPBEN|
					RCC_APB2ENR_SPI1EN;

		*this={
				{GPIOA, 5},		// spi_sck
				{GPIOA, 6},		// spi_miso (SDA in)
				{GPIOA, 7},		// spi_mosi (SDA out)
				{GPIOA, 4},		// nss (CS)
				{GPIOB, 7},		// res (hardware reset, RES)
				{GPIOB, 6}		// dcx (RS)
		};

		spi_sck.config(ALT_PP_OUT_HS);
		spi_miso.config(IN_FLOATING);
		spi_mosi.config(IN_ANALOG);
		nss.config(MAIN_PP_OUT_HS);
		res.config(MAIN_PP_OUT_HS);
		dcx.config(MAIN_PP_OUT_HS);

		SPI1->CR1=
				SPI_CR1_MSTR|
				SPI_CR1_SSM|
				SPI_CR1_SSI|
				(1<<3);				//    SCK
				//(3<<3);			//     16

		SPI1->CR1|=SPI_CR1_SPE;

		//        
		//  DMA.
		RCC->AHBENR|=RCC_AHBENR_DMA1EN;
		//  SPI1_TX   3 DMA1,
		//     :
		// MEM2MEM=0 ( "Memory to memory" );
		// PL[1:0]=00 ( );
		// MSIZE[1:0]=00 (     1 );
		// PSIZE[1:0]=01 (   ,
		// .. SPI_DR;  16 );
		// MINC=1 (    );
		// PINC=0 (    );
		// CIRC=0 ("Circular mode" );
		// DIR=1 ( -    );
		//  4   0 (   DMA
		//  ,    ).
		DMA1_Channel3->CCR=DMA1_Channel3->CCR&~0x00007FFF|
				DMA_CCR3_DIR|
				DMA_CCR3_MINC|
				DMA_CCR3_PSIZE_0;
		DMA1_Channel3->CPAR=(uint32_t)&SPI1->DR;
	}

	//   MOSI  z- (   
	// true)     push-pull (false).
	void set_mosi_z_state(bool z)
	{
		int cfg=(z)?IN_ANALOG:ALT_PP_OUT_HS;
		spi_mosi.config(cfg);
	}

	// ,    MOSI  z-.
	bool get_mosi_z_state()
	{
		return spi_mosi.get_config()==0;
	}

	//    LCD  SPI (  TXE  ,
	//  ,   ).
	void spi_send(uint8_t b)
	{
		while(!(SPI1->SR&SPI_SR_TXE))
		{}
		SPI1->DR=b;
	}

	//       SPI
	// (     
	//   LCD).
	void spi_sync()
	{
		while(
				!(SPI1->SR&SPI_SR_TXE)||
				(SPI1->SR&SPI_SR_BSY))
		{}
	}

	//   LCD   (RES).
	void reset()
	{
		set_mosi_z_state(true);
		nss.set(1);
		res.set(1);
		udelay(20000);
		res.set(0);
		udelay(20);
		res.set(1);
		udelay(20000);
		nss.set(0);
		set_mosi_z_state(false);
	}

	//   LCD    .
	void cmd_mode()
	{
		if(dcx.read_out())
		{
			spi_sync();
			dcx.set(0);
		}
	}

	//   LCD    .
	void data_write_mode()
	{
		if(!dcx.read_out())
		{
			spi_sync();
			dcx.set(1);
		}
	}

	//    LCD   DMA.
	void send_via_dma(const void *buf, uint32_t size)
	{
		if(size==0)
			return;
		DMA1_Channel3->CCR&=~1;
		DMA1->IFCR=DMA_IFCR_CGIF3;
		DMA1_Channel3->CMAR=(uint32_t)buf;
		DMA1_Channel3->CNDTR=size;
		DMA1_Channel3->CCR|=1;
		SPI1->CR2|=SPI_CR2_TXDMAEN;
		while(!(DMA1->ISR&DMA_ISR_TCIF3))
		{}
		SPI1->CR2&=~SPI_CR2_TXDMAEN;
	}
};

//         .
class Display
{
private:
	//    .
	DisplayInterface di;
public:
	//    (   
	// ).
	int x;
	int y;

	//      (    
	//  RGB).
	int color;
	int bg_color;

	// ,    LCD 
	//     RAM.
	bool f_ram_write_mode;

	//  :    
	//     .
	void init()
	{
		x=0; y=0;
		color=COLOR_WHITE;
		bg_color=0;
		f_ram_write_mode=false;

		di.init();
	}

	//   LCD     
	//    .
	void hw_reset()
	{
		x=0; y=0;
		color=COLOR_WHITE;
		bg_color=0;
		f_ram_write_mode=false;
		di.reset();
	}

	//    LCD.
	void cmd(uint8_t c)
	{
		f_ram_write_mode=false;
		di.cmd_mode();
		di.spi_send(c);
	}

	//    LCD.
	void write_data(const void *data, uint32_t size)
	{
		di.data_write_mode();
		const uint8_t *p=(const uint8_t *)data;
		for(uint32_t i=0; i<size; i++, p++)
			di.spi_send(*p);
	}

	//  LCD      .
	void display_sleep_out(int state)
	{
		state&=1;
		cmd(0x10+state);
		udelay(120000);
	}

	//    LCD
	// (      ).
	void display_on(int state)
	{
		state&=1;
		cmd(0x28+state);
	}

	//   ,    
	// ()     .
	void set_area(int x1=X_START, int x2=X_STOP, int y1=Y_START, int y2=Y_STOP)
	{
		uint8_t buf[]={0, 0, 0, 0};

		cmd(0x2A);
		buf[1]=x1;
		buf[3]=x2;
		write_data(buf, 4);

		cmd(0x2B);
		buf[1]=y1;
		buf[3]=y2;
		write_data(buf, 4);
	}

	//  LCD      RAM.
	//    ,    
	//  RAM,      
	//   (      
	//  RGB ;   6   - 
	// 18-).
	void write_ram_mode()
	{
		cmd(0x2C);
		f_ram_write_mode=true;
	}

	//    RAM ( ,   
	//     RAM).
	//     
	//  .
	void write_ram(const void *buf, uint32_t size)
	{
		if(!f_ram_write_mode)
			write_ram_mode();
		write_data(buf, size);
	}

	//      ( 
	//    ).
	inline void pixel(int _color)
	{
		write_ram(&_color, 3);
	}

	//    .
	void fill_rect(int x1, int x2, int y1, int y2, int _color)
	{
		set_area(x1, x2, y1, y2);
		int n=(x2-x1+1)*(y2-y1+1);
		for(int i=0; i<n; i++)
			pixel(_color);
	}

	//   (   ).
	void clear_screen()
	{
		fill_rect(X_START, X_STOP, Y_START, Y_STOP, bg_color);
	}

	//   (     
	//  ;     x
	//    ).
	void text_out(int ch)
	{
		ch&=0x7F;
		if(ch<0x20)
			ch=0x20;

		if(x>X_STOP||y>=Y_STOP||
				x+8<=X_START||y+8<=Y_START)
		{
			x+=8;
			return;
		}
		int xa1=(x<X_START)?X_START:x;
		int xa2=x+7;
		if(xa2>X_STOP)
			xa2=X_STOP;
		int ya1=(y<Y_START)?Y_START:y;
		int ya2=y+7;
		if(ya2>Y_STOP)
			ya2=Y_STOP;

		set_area(xa1, xa2, ya1, ya2);

		const uint8_t *p=f8x8_table[ch];
		for(int i=0; i<8; i++, p++)
		{
			int ty=y+i;
			if(ty<Y_START||ty>Y_STOP)
				continue;
			int d=*p;
			for(int j=0; j<8; j++, d<<=1)
			{
				int tx=x+j;
				if(tx<X_START||tx>X_STOP)
					continue;
				if(d&0x80)
					pixel(color);
				else
					pixel(bg_color);
			}
		}
		x+=8;
	}

	//       
	// (    ,  
	//   x     ).
	void text_out(const char *str, uint32_t length)
	{
		const char *pe=str+length;
		for(; str<pe; str++)
			text_out(*str);
	}

	//      
	// (   '\0').
	void text_out(const char *str)
	{
		while(*str!=0)
			text_out(*(str++));
	}

	//    '\0' ,   
	// ,     (  
	//      , 
	//      8     
	//  8  ).
	void print(int row, int col, const char *str)
	{
		x=col*8+X_START;
		y=row*8+Y_START;
		text_out(str);
	}

	//       ,
	//     (   
	//   ).
	void blank(int row, int col, uint32_t length)
	{
		x=col*8+X_START;
		y=row*8+Y_START;
		for(uint32_t i=0; i<length; i++)
			text_out(' ');
	}

	//      RAM 
	// (  DMA).
	// img -   ,    
	//  ( 3      RGB).
	// w, h -    ;   
	//     (  
	//   w   ,    
	//       ..).
	//        
	//   ;
	//       
	//     .
	void img_out(const void *img, uint32_t w, uint32_t h)
	{
		//   x  ,  
		//    .
		int xi0=0;

		//     (
		//  ,     ).
		int we=w;

		//        .
		int xa0=x;

		if(x<X_START)
		{
			int d=X_START-x;
			xi0=d;
			we-=d;
			xa0=X_START;
		}

		int xi1=w-1;

		if(x+w-1>X_STOP)
		{
			int d=x+w-1-X_STOP;
			xi1-=d;
			we-=d;
		}

		if(xi1<=xi0)	//  .
		{
			x+=w;
			y+=h;
			return;
		}

		//       
		//       (
		//  ).

		int yi0=0;
		int he=h;
		int ya0=y;

		if(y<Y_START)
		{
			int d=Y_START-y;
			yi0=d;
			he-=d;
			ya0=Y_START;
		}

		int yi1=h-1;

		if(y+h-1>Y_STOP)
		{
			int d=y+h-1-Y_STOP;
			yi1-=d;
			he-=d;
		}

		x+=w; y+=h;

		if(yi1<=yi0)
			return;

		set_area(xa0, xa0+we-1, ya0, ya0+he-1);
		write_ram_mode();
		di.data_write_mode();

		const uint8_t *pd=(const uint8_t*)img;
		pd+=3*(w*yi0+xi0);
		for(int i=yi0; i<=yi1; i++, pd+=3*w)
			di.send_via_dma(pd, 3*we);
	}
};

//         RAM
// (  RAM      
//   ,   
//   ;     
//   ;   - 
//       
//         ).
class Img
{
private:
	uint32_t w;		//  .
	uint32_t h;		//  .
	uint8_t *buf;	//   .
public:
	//  ,   .
	uint32_t x;
	uint32_t y;

	//       .
	int color;
	int bg_color;

	//      ;
	//   reset,    .
	Img():w(0), h(0), buf(0), x(0), y(0),
			color(COLOR_WHITE), bg_color(0) {}

	~Img()
	{
		delete[] buf;
	}

	// ,     .
	bool empty() const
	{
		return buf==0;
	}

	//   .
	inline uint32_t get_width() const
	{
		return w;
	}

	//   .
	inline uint32_t get_height() const
	{
		return h;
	}

	//    ( )  
	//       .
	void clear()
	{
		w=0;
		h=0;
		x=0;
		y=0;
		color=COLOR_WHITE;
		bg_color=COLOR_BLACK;
		delete[] buf;
		buf=0;
	}

	//        
	//    .
	void reset(uint32_t width, uint32_t height)
	{
		clear();
		buf=new uint8_t[3*width*height];
		w=width;
		h=height;
	}

	//      .
	void *get_buf()
	{
		return buf;
	}

	//       (  
	//      );
	//      0.
	void erase()
	{
		if(buf==0)
			return;
		int n=3*w*h;
		const uint8_t *pc=(const uint8_t *)&bg_color;
		uint8_t *pb=buf;
		int j=0;
		for(uint32_t i=0; i<n; i++, pb++)
		{
			if(j>=3)
				j=0;
			*pb=pc[j++];
		}

		x=0;
		y=0;
	}

	//       .
	void pixel(int _x, int _y, int _color)
	{
		if(_x>=0&&_x<w&&
				_y>=0&&_y<h)
		{
			uint8_t *p=buf+3*(_x+w*_y);
			*(p++)=_color&0xFF;
			_color>>=8;
			*(p++)=_color&0xFF;
			_color>>=8;
			*p=_color&0xFF;
		}
	}

	//   (   ).
	void text_out(char ch)
	{
		const uint8_t *p=f8x8_table[ch];
		for(int i=0; i<8; i++, p++)
		{
			int ty=y+i;
			if(ty<0||ty>=h)
				continue;
			int d=*p;
			for(int j=0; j<8; j++, d<<=1)
			{
				int tx=x+j;
				if(tx<0||tx>=w)
					continue;
				if(d&0x80)
					pixel(tx, ty, color);
				else
					pixel(tx, ty, bg_color);
			}
		}
		x+=8;
	}

	//    .
	void text_out(const char *str, uint32_t length)
	{
		const char *pe=str+length;
		for(; str<pe; str++)
			text_out(*str);
	}

	//  ,   '\0'.
	void text_out(const char *str)
	{
		while(*str!=0)
			text_out(*(str++));
	}
};
